
//////////////////////////////////////////////////////////////////////////////////////////
//											//
// file:	cracklinux.c								//
// version:	0.50									//
// author:	copyright (c) 2006 by Marko Euth <linux@euth.de>, www.euth.de		//
// last change: 09.03.2006								//
// description: This is the source code of the kernel module cracklinux.ko		//
//											//
// NOTE:	The function 'crack_DirectIO()' in this source file contains a		//
//		modified version of 'sys_ioperm()' and 'set_bitmap()', written by	//
//		obz, with changes by Linus Torvalds. The original source code can	//
//		be found in 'arch/i386/kernel/ioport.c' of the 2.6 kernel sources.	//
//											//
// license:	This program is free software; you can redistribute it and/or		//
//		modify it under the terms of the GNU General Public License		//
//		as published by the Free Software Foundation; either version 2		//
//		of the License, or (at your option) any later version.			//
//											//
//		This program is distributed in the hope that it will be useful,		//
//		but WITHOUT ANY WARRANTY; without even the implied warranty of		//
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		//
//		GNU General Public License for more details.				//
//											//
//		You should have received a copy of the GNU General Public License	//
//		along with this program; if not, write to the Free Software		//
//		Foundation, Inc., 51 Franklin Street, Fifth Floor, 			//
//		Boston, MA  02110-1301, USA						//
//											//
//////////////////////////////////////////////////////////////////////////////////////////



#include <linux/module.h>	// macros (MODULE_xxx, THIS_MODULE)
#include <linux/fs.h>		// IO Major/Minor, ioctl()
#include <asm/io.h>		// inb(), outb(), phys_to_virt(), read..(), write..()
#include <asm/uaccess.h>	// copy_from_user(), __put_user(), __get_user()
#include <linux/interrupt.h>	// request_irq(), free_irq()
#include <asm/desc.h>		// gdt table

#include "cracklinux.h"

// set module values defined in cracklinux.h that will be returned on executing modinfo
MODULE_AUTHOR(crack_AUTHOR);
MODULE_LICENSE(crack_LICENCE);
MODULE_DESCRIPTION(crack_DESCRIPTION);

// isr that's installed in kernel mode and triggers user isr
static irqreturn_t crack_INTERRUPT(int , void *, struct pt_regs *);
// enables or disables direct port access
static int crack_DirectIO(unsigned long, int);
// returns pointer on tss_struct of current cpu
static struct tss_struct * crack_GetPointerStructTSS(void);

// major number allocated by cracklinux
static int crackMAJOR = 0;

// array of waitqueues, one for each thread (one for each isr)
static wait_queue_head_t crack_WAITQUEUE[32];
// array to remember which interrupts have been installed previously
static int used_interrupts[32];

// base address of IO Permission Bitmap that's created by cracklinux
unsigned long *crack_IO_Bitmap = 0;
// pointer on thread_struct of user application
struct thread_struct * t = 0;

////////////////////////////////////////////////////////////////////////////////////////
//// open and close (executed at accessing device file) ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

// function is executed when opening the device file
static int cracklinux_open( struct inode *devicefile, struct file *instanz )
{
	#ifdef crackDEBUG2
	 printk("cracklinux: driver_open called\n");
	#endif

	// allocate memory for IO Permission Bitmap
	crack_IO_Bitmap = kmalloc(IO_BITMAP_BYTES, GFP_KERNEL);
		if (!crack_IO_Bitmap) return -cERR_NOMEM;

	// initialize bitmap
	memset(crack_IO_Bitmap, 0xFF, IO_BITMAP_BYTES);

	#ifdef crackDEBUG
	 printk("cracklinux:   allocated %d Bytes for IO Permission Bitmap\n", \
								 IO_BITMAP_BYTES);
	#endif

	// get pointer on thread_struct of opening task
	t = &current->thread;

	return 0;
}

// function is executed when closing the device file
static int cracklinux_close( struct inode *devicefile, struct file *instanz )
{
	int counter;

	#ifdef crackDEBUG2
	 printk("cracklinux: driver_close called\n");
	#endif

	// reset bitmap pointer and bitmap size to zero when kernel module
	// IO Permission Bitmap has been used previously
	if (t->io_bitmap_ptr == crack_IO_Bitmap)
	{
		#ifdef crackDEBUG
		 printk("cracklinux:   IO Bitmap has been used, reset pointer\n");
		#endif

		// disable preemption
		preempt_disable();

		// reset pointer and bitmap size
		t->io_bitmap_ptr = 0;
		t->io_bitmap_max = 0;

		// re-enable preemption
		preempt_enable();
	}

	// free space allocated for kernel module IO Permission Bitmap
	kfree(crack_IO_Bitmap);

	// remove all installed ISR when closing device file, this is only needed 
	// when irqs are not freed with CrackRemoveInterruptCallbackRoutine() or
	// CrackRemoveAllInterruptCallbackRoutines() before
	for (counter = 0; counter < 32; counter++)
	{
		// check for installed isr
		if (used_interrupts[counter] == USED)
		{
			// if installed, uninstall kernel mode isr
			free_irq(counter, &crackMAJOR);

			#ifdef crackDEBUG
			 printk("cracklinux:   irq %d is freed\n", (int) counter);
			#endif

			// mark irq as freed
			used_interrupts[counter] = FREE;

			// trigger user thread once to finish write access,
			// user thread will NOT exit on this
			wake_up_interruptible(&crack_WAITQUEUE[counter]);
		}
	}

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////
//// cracklinux_write (reads from libcracklinux) ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

// this function is called when writing on the device file (at interrupts)
static ssize_t cracklinux_write( struct file *instanz, char __user *user, \
							size_t count, loff_t *offset )
{
	int to_copy, not_copied;
	int irq;

	// copy IRQ number from library to local variable irq, necessary because
	// kernel module has to know in which waitqueue it should suspend
	to_copy = sizeof(irq);
	not_copied = copy_from_user(&irq,user,to_copy);
	if( not_copied != 0 )
		{	printk("cracklinux: ASYNCHRONOUS: DATA ERROR"); 
			return -1;
		}

	#ifdef crackDEBUG
	 printk("cracklinux: ASYNCHRONOUS: wait for interrupt %d\n", irq);
	#endif

	// sleeping till receiving interrupt
	interruptible_sleep_on(&crack_WAITQUEUE[irq]);

	#ifdef crackDEBUG
	 printk("cracklinux: ASYNCHRONOUS: trigger user thread\n");
	#endif

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////
//// IO Control ////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

// receives value transmitted from function CrackSetMode() in library,
// performs the specified operations
static int cracklinux_ioctl( struct inode *inode, struct file *instanz, \
						unsigned int cmd, unsigned long arg)
{
	PARAMETER_IO * parameter = (PARAMETER_IO *) arg;

	unsigned long physical_address;
	unsigned long *virtual_address = 0;
	unsigned long value = 0;
	unsigned long address;
	int irq;
	int status = 0;
	int type;
	int priority;

	int errCODE;

	// check whether data area is read- and writable
	if( !access_ok(VERIFY_READ,(void *)parameter,sizeof(PARAMETER_IO)) )
		return -cERR_INVAL_POINTER;
	if( !access_ok(VERIFY_WRITE,(void *)parameter,sizeof(PARAMETER_IO)) )
		return -cERR_INVAL_POINTER;

	// select function
	switch( cmd ) {
	////////////////////////////////////////////////////////////////////////////////
		case IO_READ:		// Read IO Port

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: IO Read Mode\n");
			#endif

			// get physical address and type
			__get_user(physical_address, &parameter->physical_address); 
			__get_user(type, &parameter->type);

			// read value from port
			if(type == crackBYTE){
				value = (unsigned long) inb((int) physical_address);
			}
			if(type == crackWORD){
				value = (unsigned long) inw((int) physical_address);
			}
			if(type == crackDWORD){
				value = (unsigned long) inl((int) physical_address);
			}

			#ifdef crackDEBUG
			 printk("cracklinux:   read %lX from address 0x%lX\n", \
							value, physical_address);
			#endif

			// copy value to library
			__put_user(value, &parameter->value);

			break;
	////////////////////////////////////////////////////////////////////////////////
		case IO_WRITE:		// Write IO Port

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: IO Write Mode\n");
			#endif

			// get physical address, value and type
			__get_user(physical_address, &parameter->physical_address); 
			__get_user(value, &parameter->value);
			__get_user(type, &parameter->type);

			#ifdef crackDEBUG
			 printk("cracklinux:   write data %lX to address 0x%lX\n", \
							value, physical_address);
			#endif

			// write value to port
			if(type == crackBYTE){
				outb( (int) value, (int) physical_address );
			}
			if(type == crackWORD){
				outw( (int) value, (int) physical_address );
			}
			if(type == crackDWORD){
				outl( (int) value, (int) physical_address );
			}

			break;
	////////////////////////////////////////////////////////////////////////////////
		case MEM_READ:		// Read MEM

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: MEM Read Mode\n");
			#endif

			// get physical address and type
			__get_user(physical_address, &parameter->physical_address); 
			__get_user(type, &parameter->type);

			// convert physical address to virtual address
			virtual_address = (unsigned long *) \
						phys_to_virt(physical_address);

			#ifdef crackDEBUG
			 printk("cracklinux:   physical address 0x%lX\n", \
						physical_address);
			 printk("cracklinux:   virtual address  0x%lX\n", \
						(unsigned long) virtual_address);
			#endif

			// read from determined address
			if(type == crackBYTE){
				value = (unsigned long) *(unsigned char *) \
								virtual_address;
			}
			if(type == crackWORD){
				value = (unsigned long) *(unsigned short *) \
								virtual_address;
			}
			if(type == crackDWORD){
				value = *virtual_address;
			}

			#ifdef crackDEBUG
			 printk("cracklinux:   read value 0x%lX\n", value ); 
			#endif

			// copy value to library
			__put_user(value, &parameter->value);

			break;
		/////////////////////////////////////////////////////////////////////////////
		case MEM_WRITE:		// Write MEM

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: MEM Write Mode\n");
			#endif

			// get physical address, value and type
			__get_user(physical_address, &parameter->physical_address); 
			__get_user(value, &parameter->value);
			__get_user(type, &parameter->type);

			// convert physical address to virtual address
			virtual_address = (unsigned long *) phys_to_virt(physical_address);

			#ifdef crackDEBUG
			 printk("cracklinux:   physical address 0x%lX\n", \
						physical_address);
			 printk("cracklinux:   virtual address  0x%lX\n", \
						(unsigned long) virtual_address);
			#endif

			// write value to determined address
			if(type == crackBYTE){
				* (unsigned char *) virtual_address = (unsigned char) \
									value;
			}
			if(type == crackWORD){
				* (unsigned short *) virtual_address = (unsigned short) \
									value;
			}
			if(type == crackDWORD){
				* virtual_address = value;
			}

			#ifdef crackDEBUG
			 printk("cracklinux:   written value 0x%lX\n", value ); 
			#endif

			break;
		/////////////////////////////////////////////////////////////////////////////
		case IRQ_REQUEST:	// Register IRQ, install Kenel Mode ISR

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: IRQ Request Mode\n");
			#endif

			// get irq number
			__get_user(irq, &parameter->irq); 

			#ifdef crackDEBUG
			 printk("cracklinux:   irq to request is %d\n", irq);
			#endif

			// check whether interrupt is already used by cracklinux
			if( used_interrupts[irq] == USED ){ return -cERR_ALREADYINSTALLED; }

			#ifdef crackDEBUG
			 printk("cracklinux:   irq is unused from cracklinux\n");
			#endif

			// request interrupt, install shared isr
			if( request_irq(irq, (void *) crack_INTERRUPT, SA_SHIRQ, \
				"cracklinux", &crackMAJOR) != 0)
					{ return -cERR_IRQ; }

			#ifdef crackDEBUG
			 printk("cracklinux:   irq is requested\n");
			#endif

			// remember that isr was installed
			used_interrupts[irq] = USED;

			break;
		/////////////////////////////////////////////////////////////////////////////
		case IRQ_FREE:		// Free IRQ, uninstall Kernel Mode ISR

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: IRQ Free Mode\n");
			#endif

			// get irq number
			__get_user(irq, &parameter->irq); 

			#ifdef crackDEBUG
			 printk("cracklinux:   irq to free is %d\n", irq);
			#endif

			// check for installed isr
			if (used_interrupts[irq] == USED)
			{
				// if installed, uninstall kernel mode isr
				free_irq(irq, &crackMAJOR);

				#ifdef crackDEBUG
				 printk("cracklinux:   irq %d is freed\n", irq);
				#endif

				// mark irq as freed
				used_interrupts[irq] = FREE;

				// trigger user thread once to exit it
				wake_up_interruptible(&crack_WAITQUEUE[irq]);
			}

			break;
		/////////////////////////////////////////////////////////////////////////////
		case IRQ_CHECK:		// Check for installed IRQ

			#ifdef crackDEBUG2
			 printk("cracklinux: MODE CHANGED TO: Check IRQ Mode\n");
			#endif

			// get irq number
			__get_user(irq, &parameter->irq); 

			// check whether isr is installed for this irq
			if ( used_interrupts[irq] == FREE ){ status = FREE; }
			if ( used_interrupts[irq] == USED ){ status = USED; }

			#ifdef crackDEBUG2
			 printk("cracklinux:   status is %d\n",status);
			#endif

			// copy status to library
			__put_user(status, &parameter->status);

			break;
		/////////////////////////////////////////////////////////////////////////////
		case IO_ENABLE:		// Enable Direct IO Access

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: IO Enable Mode\n");
			#endif

			// get address
			__get_user(address, &parameter->address); 

			// unlock address
			errCODE = crack_DirectIO(address, PORT_UNLOCK);
			if (errCODE != 0){ return errCODE; }

			break;
		/////////////////////////////////////////////////////////////////////////////
		case IO_DISABLE:	// Disable Direct IO Access

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: IO Disable Mode\n");
			#endif

			// get address
			__get_user(address, &parameter->address); 

			// lock address
			errCODE = crack_DirectIO(address, PORT_LOCK);
			if (errCODE != 0){ return errCODE; }

			break;
		/////////////////////////////////////////////////////////////////////////////
		case SET_NICE:		// Set new NICE level

			#ifdef crackDEBUG
			 printk("cracklinux: MODE CHANGED TO: SET NICE Mode\n");
			#endif

			// get new priority
			__get_user(priority, &parameter->priority); 

			#ifdef crackDEBUG
			 printk("cracklinux:   current nice level: %d\n", \
								current->static_prio - 120);
			#endif

			// test whether new priority is valid
			if ((priority < -20) || (priority > 19)){
				#ifdef crackDEBUG
				 printk("cracklinux:   invalid NICE level\n");
				#endif
				return -cERR_INVAL_NICE;
			}

			// set new process priority
			current->static_prio = 120 + priority;

			#ifdef crackDEBUG
			 printk("cracklinux:   new NICE level: %d\n", \
								current->static_prio-120);
			#endif

			break;
		/////////////////////////////////////////////////////////////////////////////
		default:

			#ifdef crackDEBUG
			 printk("cracklinux: UNKNOWN COMMAND, CHANGE TO STARTUP MODE\n");
			#endif

			return -cERR_MODE;

			break;
		/////////////////////////////////////////////////////////////////////////////
	}

	return 0;
}


////////////////////////////////////////////////////////////////////////////////////////
//// used functions ////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////


// kernel mode isr that's installed with request_irq()
static irqreturn_t crack_INTERRUPT(int irq, void *dev_id, struct pt_regs *regs)
{
	// Trigger waitqueue, user process will proceed on this
	wake_up_interruptible(&crack_WAITQUEUE[irq]);

	#ifdef crackDEBUG
	 printk("cracklinux: ASYNCHRONOUS: kernel mode isr %d triggered\n", irq);
	#endif

	return IRQ_HANDLED;
}

// enable and disable direct IO access
static int crack_DirectIO(unsigned long addr, int TODO )
{
	struct tss_struct * tss = 0;

	unsigned long *to_change_ptr = 0;
	unsigned int index_long;
	unsigned int index_bit;
	unsigned long mask;

	// stop here when parameter is not valid
	if (addr > IO_BITMAP_BITS){ return -cERR_INVAL_ADDR; }

	#ifdef crackDEBUG
	 printk("cracklinux:   \"%s\" (%i) enable io access on port 0x%lX\n", \
				current->comm, current->pid, addr);
	#endif

	// get number of unsigned long and bit within unsigned long
	index_long = addr / BITS_PER_LONG;
	index_bit = addr & (BITS_PER_LONG-1);

	// create bit mask for changing unsigned long
	mask = (1 << index_bit);

	// disable preemption
	preempt_disable();

	// get pointer of type struct tss_struct
	tss = crack_GetPointerStructTSS();
	#ifdef crackDEBUG
	 printk("cracklinux:   Address of tss_struct:     0x%lX\n", (unsigned long) tss);
	#endif

	// if it's the first ioperm() call in this task's lifetime,
	// set IO bitmap up (use bitmap created in init function)
	if (t->io_bitmap_ptr == 0)
	{
		#ifdef crackDEBUG
		 printk("cracklinux:   Bitmap doesn't exist, use own...\n"); 
		#endif
		t->io_bitmap_ptr = crack_IO_Bitmap;
		t->io_bitmap_max = IO_BITMAP_LONGS;
	}

	// get pointer of unsigned long that should be changed
	to_change_ptr = t->io_bitmap_ptr + index_long;

	#ifdef crackDEBUG
	 printk("cracklinux:   long index: %u, bit index: %u, mask: 0x%lX\n", \
				index_long, index_bit, mask );
	 printk("cracklinux:   long changed from 0x%lX ", *to_change_ptr );
	#endif

	// change bit in bitmap
	if (TODO == PORT_UNLOCK){ *to_change_ptr &= ~mask; }
	if (TODO == PORT_LOCK){ *to_change_ptr |= mask; }

	#ifdef crackDEBUG
	 printk("into 0x%lX\n", *to_change_ptr );
	#endif

	// force reloading the correct bitmap at next IO operation
	tss->io_bitmap_base = INVALID_IO_BITMAP_OFFSET_LAZY;

	// re-enable preemption
	preempt_enable();

	return 0;
}

// get pointer on tss_struct of current cpu
static struct tss_struct * crack_GetPointerStructTSS(void)
{
	// cpu-id
	int cpu;

	// 8 byte segment descriptor
	struct own_desc_struct { 
		u16 limit0;
		u16 base0;
		unsigned base1 : 8, type : 4, s : 1, dpl : 2, p : 1;
		unsigned limit : 4, avl : 1, l : 1, d : 1, g : 1, base2 : 8;
	};
	struct own_desc_struct * tss_desc = 0;

	unsigned long base_addr = 0;
	char * part = 0;

	// get id of current CPU
	cpu = smp_processor_id();

	// get pointer on GDT of current cpu  (segment.h)
	tss_desc = (struct own_desc_struct *) per_cpu(cpu_gdt_table, cpu);

	#ifdef crackDEBUG
	 printk("cracklinux:   Address of GDT Table:      0x%lX\n", \
							(unsigned long) tss_desc);
	#endif

	// get pointer on TSS descriptor in GDT
	tss_desc += GDT_ENTRY_TSS;

	#ifdef crackDEBUG
	 printk("cracklinux:   Address of TSS Descriptor: 0x%lX\n", \
							(unsigned long) tss_desc);
	#endif

	// write base address to base_addr
	base_addr = (unsigned long) tss_desc->base0;	// write byte 1+2 of address
	part = (char *) &base_addr;			// write byte3 of address
	part += 2;
	*part = tss_desc->base1;
	part += 1;					// write byte4 of address
	*part = tss_desc->base2;

	// return pointer of type struct tss_struct
	return (struct tss_struct *) base_addr;
}


////////////////////////////////////////////////////////////////////////////////////////
//// function pointer for interface functions //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

// write function pointer for open/close, write and ioctl functions to structure
static struct file_operations crack_FOPS = {
	.owner   = THIS_MODULE,
	.open    = (void *) cracklinux_open, 
	.release = (void *) cracklinux_close,
	.write   = (void *) cracklinux_write,
	.ioctl   = (void *) cracklinux_ioctl,
};

////////////////////////////////////////////////////////////////////////////////////////
//// define and set init and exit functions ////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

// init function, it's executed on loading the module with insmod or modprobe
static int __init init_cracklinux(void)
{
	int counter;

	#ifdef crackDEBUG
	 printk("cracklinux: init called, loading Module...\n");
	#endif

	// initialise all 32 waitqueues of array crack_WAITQUEUE, one 
	// waitqueue is necessary for each interrupt. They've got to 
	// be initialised before first usage.
	for ( counter=0; counter < 32; counter++ ){
		init_waitqueue_head (&crack_WAITQUEUE[counter]);
	}

	// register dynamic major number, module name and previously defined
	// functions in kernel	
	crackMAJOR = register_chrdev(0, "cracklinux", &crack_FOPS);

	if(crackMAJOR < 0)
	{
		printk("cracklinux:   ERROR: unable to registry major number\n");
		return( -EIO );
	};

	#ifdef crackDEBUG
	 printk("cracklinux:   major number registered: %d\n",crackMAJOR);
	#endif

      	return 0;
}

// exit function, it's executed on unloading the module with rmmod or modprobe -r
static void __exit cleanup_cracklinux(void)
{
	#ifdef crackDEBUG
	 printk("cracklinux: exit called, unloading Module...\n");
	#endif

	// free major number in kernel
	if( unregister_chrdev(crackMAJOR, "cracklinux") == 0 )
	{
		#ifdef crackDEBUG
		 printk("cracklinux:   major number freed.\n");
		#endif
	}

	#ifdef crackDEBUG
	 printk("cracklinux:   remove Module...\n");
	#endif
}

// tell init and exit functions to module (because their names are arbitrary from user)
module_init(init_cracklinux);
module_exit(cleanup_cracklinux);

////////////////////////////////////////////////////////////////////////////////////////
//// EOF ///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
